/*
 * Copyright 2021 Karl F. Meinkopf
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

import 'dart:async';
import 'dart:isolate';

import 'messaging.dart';
import 'util.dart';

typedef Filter = bool Function(Message message);

abstract class _IntercomBase {
  ReceivePort get _inPort;

  SendPort get _outPort;

  OrderedListener<dynamic> get listener;

  bool get isInitialized;

  FutureOr<void> initialize();

  void _onEvent(dynamic data) => listener.onEvent(data);

  Future<Message> sendWaitWhere(Message message, Filter test,
      {Duration? timeout}) {
    var comp = Completer<Message>();

    var lsid = listener.prependListener((e) {
      try {
        var msg = Message.fromJson(assertTyped(e));
        if (test(msg)) {
          comp.complete(msg);
          return false;
        }
      } on TypeAssertionError {
        return true;
      }

      return true;
    });

    var future = comp.future.then((value) {
      listener.removeListener(lsid);
      return value;
    });

    sendSignal(message);

    return (timeout != null) ? future.timeout(timeout) : future;
  }

  void sendSignal(Message message) => sendRaw(message.toJson());

  void sendRaw(data) => _outPort.send(data);

  void close() {
    _inPort.close();
  }
}

class ManagerIntercom extends _IntercomBase {
  late final Uri uri;
  late final Isolate isolate;
  @override
  late final ReceivePort _inPort;
  @override
  late final SendPort _outPort;
  @override
  final OrderedListener<dynamic> listener = OrderedListener();
  bool _initialized = false;

  ManagerIntercom(this.uri);

  @override
  bool get isInitialized => _initialized;

  @override
  Future<void> initialize(
      {List<String> args = const [],
      Duration timeout = const Duration(seconds: 2)}) async {
    var comp = Completer();

    _inPort = ReceivePort();

    isolate = await Isolate.spawnUri(uri, args, _inPort.sendPort);

    var lsid = listener.prependListener((message) {
      if (!_initialized) {
        if (message is SendPort) {
          _initialized = true;
          _outPort = message;

          comp.complete();
          return false;
        }
      }
      return true;
    });

    _inPort.listen(_onEvent);

    return comp.future.timeout(timeout).then((value) {
      listener.removeListener(lsid);
      return value;
    }).onError((error, stackTrace) {
      _inPort.close();
      if (error != null) throw error;
    });
  }

  @override
  void close() {
    super.close();
  }
}

class PluginIntercom extends _IntercomBase {
  @override
  final ReceivePort _inPort;
  @override
  final SendPort _outPort;
  @override
  final OrderedListener listener = OrderedListener();
  bool _initialized = false;

  PluginIntercom(this._inPort, this._outPort);

  @override
  bool get isInitialized => _initialized;

  @override
  void initialize() {
    _initialized = true;
    _inPort.listen(_onEvent);
    _outPort.send(_inPort.sendPort);
  }
}

abstract class Filters {
  static Filter ofType(List<MessageType> types) =>
      (Message msg) => types.contains(msg.type);

  static Filter every(List<Filter> filters) =>
      (Message msg) => filters.every((filter) => filter(msg));

  static Filter any(List<Filter> filters) =>
      (Message msg) => filters.any((filter) => filter(msg));
}
